; ****************************************************************************
;                 AT90S2313 MIDI CONTROLLER FOR THE SN76469 
; ****************************************************************************
;
;
; Copyright 2009 Alexis Kotlowy. Released under the GPL3.
;
; http://www.gnu.org/licenses/gpl-3.0-standalone.html
;
; This will assemble with avra. Available at avra.sourceforge.net
;
; email thrashbarg@kaput.homeunix.org if you have any questions.
;
; NO WARRANTY!
;

		.device AT90S2313 
		.include "2313.def"


; Channel equate. Determines what MIDI channel is interpreted.
; Must be between 0 and 15.
		.equ CHANNEL	= 0		; MIDI channel
 
; Decay rate while note is playing is n times slower than the release
; rate.
		.equ DECAY	= 10
; Release rate. This value is written directly to the timer. Lower is
; quicker.
		.equ RELEASE	= 5

; Register Defines
		
		.def TEMP	= r16		; General Purpose Register
		.def TEMPB	= r17		; Used to store SRAM data
		.def TONEH	= r18		; High tone byte from EEPROM
		.def TONEL	= r19		; Low tone byte
		.def RX_TEMP	= r20		; Used in Interrupt Routine
		.def CURRENT	= r21		; Currently used channel
		.def COUNT	= r22		; Decay timer counter
		.def VEL	= r23		; Velocity

; Index Register Defines

		.def XL		= r26		; FIFO In pointer
		.def XH		= r27
		.def YL		= r28		; FIFO Out pointer
		.def YH		= r29
		.def ZL		= r30		; General Index Register
		.def ZH		= r31

; SRAM locations
		.equ KEY1	= 0xA0		; Current key on Channel 1
		.equ KEY2	= 0xA1		; Current key on Channel 2
		.equ KEY3	= 0xA2		; Current key on Channel 3
		.equ ATTEN1	= 0xA3		; Attenuation on Channel 1
		.equ ATTEN2	= 0xA4		; Attenuation on Channel 2
		.equ ATTEN3	= 0xA5		; Attenuation on Channel 3

; SN76496 Register Equates
		
		.equ NOTE_F	= 0b10000000	; Note Frequency
		.equ NOTE_A	= 0b10010000	; Note Attenuation

		.equ CHAN1	= 0x00		; Channel 1
		.equ CHAN2	= 0x20		; Channel 2
		.equ CHAN3	= 0x40		; Channel 3
		.equ NOISE	= 0x60		; Noise
		
; FIFO Location Equate
		.equ FIFO	= 0x60		; Start
		.equ FIFO_END	= 0x9F		; End

; Code Starts
	
		.cseg
		.org	0

; Interrupt Vector Table
		rjmp	RESET			; Reset Vector
		nop				; External Interrupt 0
		nop				; External Interrupt 1
		nop				; Timer 1 Capture
		nop				; Timer 1 Compare Match
		nop				; Timer 1 Overflow
		rjmp	TIM_OVF0		; Timer 0 Overflow
		rjmp	UART_RX			; UART RX
		nop				; UART Data Empty
		nop				; UART TX
		nop				; Analog Comparator

; Main Program
;
MAIN:		rcall	FIFO_READ		; Get command from FIFO
		
CMDS:		cpi	TEMP, 0x80 + CHANNEL	; Release?
		breq	NOTE_OFF
		cpi	TEMP, 0x90 + CHANNEL 	; Strike?
		breq	NOTE_ON
		cpi	TEMP, 0xFF		; System Reset
		breq	RESET_I

		rjmp	MAIN			; Unknown? Then loop

 RESET_I:	rjmp	RESET			; Cursed offsets!

; Jump back to here to get the Velocity operand
NOTE_OFF_VEL:	rcall	FIFO_READ		; Result is ignored

; Disable note.
NOTE_OFF:	rcall	FIFO_READ		; Note

		sbrc	TEMP, 7			; Highest bit clear?
		rjmp	CMDS			; No? Process a command

		lds	TEMPB, KEY1		; Status of KEY1?
		cp	TEMP, TEMPB		; Key off for Channel 1?
		breq	NOTE_OFF_1		; Yes? Jump
		lds	TEMPB, KEY2		; Channel 2?
		cp	TEMP, TEMPB		; ...
		breq	NOTE_OFF_2		; Yes? Jump
		lds	TEMPB, KEY3		; Channel 3
		cp	TEMP, TEMPB		; ...
		breq	NOTE_OFF_3		; Yes? Jump
		rjmp	NOTE_OFF_VEL		; Continue reading off events

NOTE_OFF_1:	ldi	TEMP, 0xFF		; Flag key off in RAM
		sts	KEY1, TEMP		; ...
		lds	TEMP, ATTEN1		; Get current attenuation
		ori	TEMP, 0x80		; Flag release
		sts	ATTEN1, TEMP		; Store to register
		rjmp	NOTE_OFF_VEL		; Process Velocity event

NOTE_OFF_2:	ldi	TEMP, 0xFF		; Flag key off in RAM
		sts	KEY2, TEMP		; ...		
		lds	TEMP, ATTEN2		; Get current attenuation
		ori	TEMP, 0x80		; Flag release
		sts	ATTEN2, TEMP		; Store to register
		rjmp	NOTE_OFF_VEL		; Process Velocity event

NOTE_OFF_3:	ldi	TEMP, 0xFF		; Flag key off in RAM
		sts	KEY3, TEMP		; ...		
		lds	TEMP, ATTEN3		; Get current attenuation
		ori	TEMP, 0x80		; Flag release
		sts	ATTEN3, TEMP		; Store to register
		rjmp	NOTE_OFF_VEL		; Process Velocity event


; Respond to key press.
; Send frequency from frequency table to SN76496,
; Turn note on.
NOTE_ON:	rcall	FIFO_READ		; Get Note

		sbrc	TEMP, 7			; Highest bit clear?
		rjmp	CMDS			; No? Process a command

		; Get velocity and convert to an attenuation level

		push	TEMP			; Keep for attenuation

		rcall	FIFO_READ		; Get velocity
		com	TEMP			; Attenuation, not velocity
		lsr	TEMP			; Move left three bytes
		lsr	TEMP			; ...
		lsr	TEMP			; ...
		andi	TEMP, 0x0F		; Get rid of any rubbish
		mov	VEL, TEMP		; To use later
		
		pop	TEMP

		; Scan through the channel list to find
		; an unused channel.

		lds	TEMPB, KEY1		; Check Channel 1
		sbrs	TEMPB, 7		; Highest bit set?
		rjmp	NOTE_ON_2		; No? Jump

		ldi	CURRENT, CHAN1		; Enable Channel 1
		sts	KEY1, TEMP		; Store note to channel status
		sts	ATTEN1, VEL		; Store to RAM
		rjmp	NOTE_ON_FREQ		; Continue

NOTE_ON_2:	lds	TEMPB, KEY2		; Check Channel 2
		sbrs	TEMPB, 7		; Highest bit set?
		rjmp	NOTE_ON_3		; No? Jump
	
		ldi	CURRENT, CHAN2		; Enable Channel 2
		sts	KEY2, TEMP		; Store note to channel status
		sts	ATTEN2, VEL		; Store to RAM
		rjmp	NOTE_ON_FREQ		; Continue

NOTE_ON_3:	lds	TEMPB, KEY3		; Check Channel 3
		sbrs	TEMPB, 7		; Highest bit set?
		rjmp	MAIN			; No? All channels full
	
		ldi	CURRENT, CHAN3		; Enable Channel 3
		sts	KEY3, TEMP		; Store note to channel status
		sts	ATTEN3, VEL		; Store to RAM
		;rjmp	NOTE_ON_FREQ		; Fall through

		; Send frequency from EEPROM to PSG

NOTE_ON_FREQ:	subi	TEMP, 36		; Data starts at 36th note
		out	EEAR, TEMP		; Set address
		sbi	EECR, 0			; Enable read
		in	TONEH, EEDR		; Get data from EEPROM
		
		subi	TEMP, -60		; move forward to second table
		out	EEAR, TEMP		; Send address to EEPROM
		sbi	EECR, 0			; Enable read
		in	TEMP, EEDR		; Get data from EEPROM

		andi	TEMP, 0x0F		; Low byte should only be here
		ori	TEMP, NOTE_F		; Combine with PSG command
		or	TEMP, CURRENT		; Combine with channel
		rcall	SENDB			; send Command to SN76496
		mov	TEMP, TONEH		; send Frequency
		andi	TEMP, 0x3F
		rcall	SENDB			; ...

		; Set correct attenuation on channel

		or	VEL, CURRENT		; Current channel
		ori	VEL, NOTE_A		; Attenuation Command
		mov	TEMP, VEL
		rcall	SENDB
		
		rjmp	NOTE_ON			; Continue interpreting notes



; Read byte from FIFO. If empty, wait
; Byte returned in TEMP
;
FIFO_READ:	cp	XL, YL			; Wait for data to enter...
		breq	FIFO_READ		; the FIFO
		
		ld	TEMP, Y+		; Get byte off FIFO, increment
		cpi	YL, FIFO_END		; Loop the FIFO around?
		brne	FIFO_R_END		; No? Jump
		ldi	YL, FIFO		; Yes? Reset
		clr	YH			; Just to be sure
FIFO_R_END:	ret



; Send Byte routine.
; Byte to send is in TEMP.
; TEMP destroyed.
; While the READY pin is connected to the INT0 pin, I couldn't get it
; to use it as an interrupt source reliably. Perhaps you can!
; The uC is controlling the PSG 'blind', as it doesn't know what state
; the READY pin is in.
;
SENDB:		out	PORTB, TEMP		; Send byte to data bus
		cbi	PORTD, 3		; Lower WE
		ldi	TEMP, 16		; Wait for a few cycles
SENDB_WAIT:	dec	TEMP			; wait
		brne	SENDB_WAIT		; wait
		sbi	PORTD, 3		; Raise WE
		ret				; Bye

; Reset Vector entry
;
RESET:		cli
		ldi	TEMP, 0			; Clear R1
		mov	r1, TEMP
		ldi	TEMP, 0b00011000	; Disable Watchdog
		out	WDTCR, TEMP
		ldi	TEMP, 0b00010000
		out	WDTCR, TEMP

		ldi	TEMP, 0xDF		; Initialize Stack Pointer
		out	SPL, TEMP

		ldi	TEMP, 0b00000000	; Data Bus all off
		out	PORTB, TEMP
		ldi	TEMP, 0b11111111	; Data Bus all outputs
		out	DDRB, TEMP

		ldi	TEMP, 0b00001100	; Control Bus WE high
		out	PORTD, TEMP
		ldi	TEMP, 0b00001000	; Control Bus WE output
		out	DDRD, TEMP

		ldi	XL, FIFO		; Initialize MIDI FIFO
		clr	XH
		mov	YL, XL
		clr	YH

		ldi	TEMP, 3			; UART rate (31250bps @ 2MHz)
		out	UBRR, TEMP
		ldi	TEMP, 0b10010000	; UART RX enable, IRQ enable
		out	UCR, TEMP

		ldi	TEMP, 0b00000100	; Timer 0 CLK/256
		out	TCCR0, TEMP
		ldi	TEMP, RELEASE		; Release rate
		out	TCNT0, TEMP

		ldi	TEMP, 0b00000010	; Enable Timer 1 overflow IRQ
		out	TIMSK, TEMP

		ldi	TEMP, 0x8F		; Full attenuation
		sts	ATTEN1, TEMP		; Channel 1
		sts	ATTEN2, TEMP		; Channel 2
		sts	ATTEN3, TEMP		; Channel 3
		ldi	TEMP, 0xFF		; Note off
		sts	KEY1, TEMP		; Channel 1
		sts	KEY2, TEMP		; Channel 2
		sts	KEY3, TEMP		; Channel 3

		ldi	COUNT, DECAY		; Reset COUNT to DECAY rate

		ldi	TEMP, NOTE_A+CHAN1+0xF	; Turn off volume to all
		rcall	SENDB
		ldi	TEMP, NOTE_A+CHAN2+0xF
		rcall	SENDB
		ldi 	TEMP, NOTE_A+CHAN3+0xF
		rcall	SENDB
		ldi 	TEMP, NOTE_A+NOISE+0xF
		rcall	SENDB

		ldi	TEMP, 0b11100100	; Initialize White Noise
		rcall	SENDB

		sei				; Enable Interrupts

		rjmp	MAIN


; UART receive Interrupt Service Routine. Send incoming data
; to FIFO, loop FIFO pointer if required.
UART_RX:	lds	RX_TEMP, SREG
		push	RX_TEMP
		in	RX_TEMP, UDR		; Get UART data
		st	X+, RX_TEMP		; Store to FIFO
		cpi	XL, FIFO_END		; Loop FIFO around?
		brne	UART_RX_X		; No? Jump
		ldi	XL, FIFO		; Yes? Reset
		clr	XH			; Just to be sure
UART_RX_X:	pop	RX_TEMP
		sts	SREG, RX_TEMP
		reti				; Bye

; Timer 0 Interrupt Service Routine
; Controls decay volume.
;
TIM_OVF0:	push	TEMP			; don't destroy this!
		lds	TEMP, SREG		; Save status register
		push	TEMP
		push	TEMPB

		dec	COUNT			; Count DECAY times per IRQ
		brne	TIM_REL			; Not ready? Jump

		ldi	COUNT, DECAY		; Reset DECAY

		lds	TEMP, ATTEN1		; Get Attenuation status
		rcall	DECAY_SUB		; Attenuation subroutine
		sts	ATTEN1, TEMP		; Update register

		lds	TEMP, ATTEN2		; Attenuation Status, ch2
		rcall	DECAY_SUB		; Run attenuation sub
		sts	ATTEN2, TEMP		; Update attenuation status

		lds	TEMP, ATTEN3		; Attenuation Status, ch3
		rcall	DECAY_SUB		; Run attenuation sub
		sts	ATTEN3, TEMP		; Update attenuation status

TIM_REL:	lds	TEMP, ATTEN1		; Get Attenuation
		rcall	RELEASE_SUB		; Process release rate
		sts	ATTEN1, TEMP		; Write changes to RAM
		ori	TEMP, NOTE_A + CHAN1	; Combine with cmd & chan
		rcall	SENDB			; Send to PSG

		lds	TEMP, ATTEN2		; Get Attenuation
		rcall	RELEASE_SUB		; Process release rate
		sts	ATTEN2, TEMP		; Write changes to RAM
		ori	TEMP, NOTE_A + CHAN2	; Combine with cmd & chan
		rcall	SENDB			; Send to PSG

		lds	TEMP, ATTEN3		; Get Attenuation
		rcall	RELEASE_SUB		; Process release rate
		sts	ATTEN3, TEMP		; Write changes to RAM
		ori	TEMP, NOTE_A + CHAN3	; Combine with cmd & chan
		rcall	SENDB			; Send to PSG

TIM_X:		pop	TEMPB
		pop	TEMP			; restore Status Reg
		sts	SREG, TEMP
		pop	TEMP
		reti				; return

RELEASE_SUB:	mov	TEMPB, TEMP		; To work on
		andi	TEMPB, 0x0F		; Mask release flag
		cpi	TEMPB, 0x0F		; Already at lowest?
		breq	RELEASE_SUB_X		; Don't lower further
		sbrc	TEMP, 7			; Don't lower if in decay
		inc	TEMP
RELEASE_SUB_X:	ret


DECAY_SUB:	mov	TEMPB, TEMP		; To work on
		andi	TEMPB, 0x0F		; Mask release flag
		cpi	TEMPB, 0x0F		; Already Lowest?
		breq	DECAY_SUB_X		; Don't decay further
		sbrs	TEMP, 7			; Don't change if released
		inc	TEMP			; Next level of attenuation
DECAY_SUB_X:	ret				; Bye

; Frequency table
; Divided into two segments, one for the 4 bits sent with the command and
; one for the second parameter.
	.eseg
	.org	0

FREQH:
;       C    C#    D     D#    E     F     F#    G     G#    A     A#    B
 .db 0x3B,0x38,0x35,0x32,0x2F,0x2C,0x2A,0x27,0x25,0x23,0x21,0x1F
 .db 0x1D,0x1C,0x1A,0x19,0x17,0x16,0x15,0x13,0x12,0x11,0x10,0x0F
 .db 0x0E,0x0E,0x0D,0x0C,0x0B,0x0B,0x0A,0x09,0x09,0x08,0x08,0x07
 .db 0x07,0x07,0x06,0x06,0x05,0x05,0x05,0x04,0x04,0x04,0x04,0x03
 .db 0x03,0x03,0x03,0x03,0x02,0x02,0x02,0x02,0x02,0x02,0x02,0x01

FREQL:
 .db 0x0B,0x05,0x03,0x03,0x06,0x0B,0x03,0x0D,0x09,0x08,0x08,0x0A
 .db 0x0D,0x02,0x09,0x01,0x0B,0x05,0x01,0x0E,0x0C,0x0C,0x0C,0x0D
 .db 0x0E,0x01,0x04,0x08,0x0D,0x02,0x08,0x0F,0x06,0x0E,0x06,0x0E
 .db 0x07,0x00,0x0A,0x04,0x0E,0x09,0x04,0x0F,0x0B,0x07,0x03,0x0F
 .db 0x0B,0x08,0x05,0x02,0x0F,0x0C,0x0A,0x07,0x05,0x03,0x01,0x0F
